from meili_sdk.websockets.client import MeiliWebsocketClient
from sentry_sdk import capture_exception, capture_message
from meili_lib.agent import agent_sentry
from meili_lib.parse_data import parse_waypoints, parse_topic, parse_docking_routine


from std_msgs.msg import Bool


class Handlers:

    def __init__(self, agent):
        self.agent = agent
        self.client = []  # docking routine

        for i in range(0, self.agent.number_of_vehicles):
            self.client.append(0)  # docking routine

        self.rate = None

    def sdk_setup(self):
        """sdk_setup"""
        try:
            self.agent.client = MeiliWebsocketClient(
                self.agent.token,
                override_host=self.agent.server_instance.replace("https", "wss"),
                fleet=self.agent.fleet,
                open_handler=self.open_handler,
                close_handler=self.close_handler,
                error_handler=self.error_handler,
                task_handler=self.task_handler,
                task_cancellation_handler=self.task_cancellation_handler,
                move_action_handler=self.move_action_handler,
                slow_down_handler=self.slow_down_handler,
                topic_list_initializer_handler=self.topic_handler,
                # path_rerouting_handler=self.path_rerouting_handler,
                # docking_routine_request_handler=self.docking_routine_request_handler,
                # docking_routine_finalize_handler=self.docking_routine_finalize_handler,
                # collision_clearance_handler=self.collision_clearance_handler,
            )
        except KeyError as error:
            self.agent.log_info(f"[ROSHandler] No handler named {error}")
            capture_exception(error)
        return self.agent.client

    def __check_fleet_get_vehicle_position(self, vehicle):
        if self.agent.fleet:
            return self.agent.return_vehicle_position(vehicle)
        return 0

    def move_action_handler(self, _, data: dict, vehicle: str):
        """When a task is received, goal is generated and sent to the specific robot"""
        try:
            vehicle_position = self.__check_fleet_get_vehicle_position(vehicle)
            self.send_goal(data, vehicle_position)
            self.rate.sleep()

            # If the robot is moving the task will be automatically cancelled to go to the
            # charging station.
            self.agent.log_info(
                f"[ROSHandler] Sending vehicle {self.agent.vehicle_list[vehicle_position]} to CHARGING STATION"
            )

        except Exception as error:
            self.agent.log_error(
                "[ROSHandler] Error move_action_handler: %s" % (error,)
            )
            self.agent.agent_sentry(error, "[ROSHandler] Error move_action_handler")

    def task_handler(self, _, data: dict, vehicle: str):
        """When a task is received, goal is generated and sent to the specific robot"""
        try:
            vehicle_position = self.__check_fleet_get_vehicle_position(vehicle)
            current_task_uuid = data["uuid"]

            self.agent.number_of_tasks = self.agent.number_of_tasks + 1

            self.agent.log_info(
                f"[ROSHandler] Number of received task is: {self.agent.number_of_tasks}"
            )
            self.agent.log_info(
                f"[ROSHandler] Received task assigned to vehicle: {self.agent.vehicle_list[vehicle_position]}"
            )
            if not data["metric_waypoints"]:
                # Case when only one task is sent
                self.send_goal(data, vehicle_position)
            else:
                # This is batch-task vehicle
                if self.agent.node.path_planning:
                    pose = parse_waypoints(data)
                    self.waypoints_handler(
                        vehicle_position=vehicle_position,
                        pose=pose,
                        current_task_uuid=current_task_uuid,
                    )
                else:
                    self.send_goal(data, vehicle_position)

            self.rate.sleep()

        except KeyError as error:
            self.agent.log_error(
                f"[ROSHandler] Error task_handler: {error}"
            )
            agent_sentry(error, "[ROSHandler] Error task_handler")

    def task_cancellation_handler(self, goal_id, vehicle):
        """Handles task cancellation info to specific robot"""
        self.agent.log_info(f"[ROSHandler] Cancelling task with goal id: {goal_id}")
        self.agent.log_info(f"[ROSHandler] Cancelling task of vehicle: {vehicle}")

        # if cancellation is received cancel goal
        vehicle_position = self.__check_fleet_get_vehicle_position(vehicle)

        self.cancel_goal(goal_id, vehicle_position)

        # Refresh variables
        self.agent.waypoints[vehicle_position] = False
        self.agent.task_started[vehicle_position] = 0

    def waypoints_handler(
            self, vehicle_position, pose, current_task_uuid
    ):
        """Handles offline waypoints and sends to specific robot"""
        # ROS2
        try:
            self.agent.waypoints[vehicle_position] = True
            self.agent.current_task_uuid = current_task_uuid

            self.send_waypoints(pose, vehicle_position)
        except KeyError as error:
            self.agent.log_error(
                f"[ROSHandler] Error waypoints_handler:{error}"
            )

            self.agent.agent_sentry(error, "[ROSHandler] Error waypoints_handler")

    #def get_list_of_topic():
    #    """Getting the list of topics available with specified namespace"""
    #    # getting  topics available
    #    topic_list = rospy.get_published_topics(namespace="/robot1")
    #    return topic_list

    def topic_request_handler(self, _):
        """logging information"""
        self.agent.log_info("[ROSHandler] Received topic request")
        # get_list_of_topic()

    def close_handler(self, *_):
        """Close the websocket"""
        self.agent.log_info("[ROSHandler] WS close for Agent")
        capture_message("[ROSHandler] WS close for Agent")

    def error_handler(self, _, err):
        """verifies websocket connection"""
        if not self.agent.offline.offline:
            self.agent.log_error(f"[ROSHandler] WS error {err}")
            capture_message("[ROSHandler] WS error")

    def open_handler(self, *_, **__):
        """Opens the websocket connection"""
        self.agent.log_info("[ROSHandler] WS open")
        self.agent.ws_open = True

    def topic_handler(self, _, topics, vehicle):
        """find vehicle token to passed to the topic streaming node"""

        if topics and vehicle:

            # initialise custom msg to send topic info

            for index in topics["data"]:
                self.agent.total_topics += 1
                index["vehicle_uuid"] = vehicle

                msg = parse_topic(index, self.agent)

                self.agent.topic_streaming_talker(msg)
                self.agent.log_info(
                    f"[ROSHandler] Topic to stream {index['topic']}, for vehicle {vehicle}"
                )
        else:
            self.agent.log_info("[ROSHandler] No topics to stream")

    def path_rerouting_handler(self, _, data: dict, vehicle: str):
        vehicle_position = self.__check_fleet_get_vehicle_position(vehicle)

        self.agent.goal_id_rerouted[vehicle_position] = self.agent.goal_id[vehicle_position]
        self.task_cancellation_handler(goal_id=self.agent.goal_id_rerouted[vehicle_position], vehicle=vehicle)

        data_new_path = {"rotation_angles": data["rotation_angles"], "metric_waypoints": data["path"]}
        current_task_uuid = "Rerouted_task"
        pose = parse_waypoints(data_new_path)
        self.waypoints_handler(
            vehicle_position=vehicle_position,
            pose=pose,
            current_task_uuid=current_task_uuid,
        )

    def docking_routine_finalize_handler(self, data: dict, vehicle: str):
        self.agent.log_info("[ROSHandler] Docking Routine FINALIZING request")
        vehicle_position = self.__check_fleet_get_vehicle_position(vehicle)

        rate = self.agent.node.create_rate(10)

        try:
            vehicle = self.agent.vehicles[vehicle_position]
            vehicle_prefix = str(vehicle["prefix"])
            if vehicle_prefix == "None":
                vehicle_prefix = ""

            pub = self.end_recording_publisher(vehicle_prefix)

            for i in range(10):
                variable = Bool()
                variable.data = False
                pub.publish(variable)
                rate.sleep()
            self.agent.log_info("[ROSHandler] Published end of recording")

            self.client[vehicle_position].wait_for_result()
            result = self.client[vehicle_position].get_result()

            if result is not None:
                # rospy.loginfo(f"[ROSHandler] Result {result}")
                path, rotation_angles = parse_docking_routine(result.plan)
                self.agent.ping_docking_routine(vehicle["uuid"], path, rotation_angles)
            else:
                self.agent.log_error(f"[ROSHandler] Docking output is None. Try again.")
        except Exception as e:
            self.agent.log_error(f"[ROSHandler] Error on Docking Routine Finalize: {e}")






from geometry_msgs.msg import PoseStamped, Twist
from meili_agent_msgs.msg import Topicinit
from meili_lib.maths import quaternion_to_euler, euler_to_quaternion

import math


def rotation_setup(rotation):
    rotation_float = float(rotation)

    if float(rotation) > 180:
        rotation_float = -360 + float(rotation)
    elif float(rotation) == 180:
        rotation_float = -180

    # rotation_quaternion = {"qx": 0, "qy": 0, "qw": 0, "qz": 0}
    rotation_quaternion = euler_to_quaternion(0, 0, math.radians(rotation_float))

    return rotation_quaternion


def goal_setup(data):
    """Set up the goal with rotation, x and y"""
    x_meters = data["location"]["metric"]["x"]
    y_meters = data["location"]["metric"]["y"]
    try:
        rotation = data["location"]["rotation"]
    except KeyError:
        rotation = 0

    rotation_quaternion = rotation_setup(rotation)

    return x_meters, y_meters, rotation_quaternion


def parse_waypoints(data):
    pose = []
    rotation_angle = data["rotation_angles"]
    waypoints = data["metric_waypoints"]

    for index, waypoint in enumerate(waypoints):
        x = waypoint[0]
        y = waypoint[1]
        rotation = rotation_angle[index]
        pose.append([x, y, rotation])
    # pose.pop(0)

    return pose


def parse_battery_data(msg):
    # creating dictionary for storing battery data
    battery_data = {}

    if hasattr(msg, "capacity"):
        battery_data["value"] = round(msg.capacity, 2)
        battery_data["power_supply_status"] = msg.power_supply_status
    elif hasattr(msg, "level"):
        battery_data["value"] = round(msg.level, 2)
        if msg.is_charging:
            battery_data["power_supply_status"] = 1
        else:
            battery_data["power_supply_status"] = 3
    elif hasattr(msg, "data"):
        battery_data["value"] = msg.data
        battery_data["power_supply_status"] = 3

    return battery_data


def parse_vel(max_vel_x):
    # TO DO: get the current parameters and just change the linear.x
    vel = Twist()

    vel.linear.x = 0.0  # float(max_vel_x)
    vel.linear.y = 0.0
    vel.linear.z = 0.0

    vel.angular.x = 0.0
    vel.angular.y = 0.0
    vel.angular.z = 0.0

    return vel


def parse_topic(data, agent):
    vehicle_position = agent.return_vehicle_position(data["vehicle_uuid"])

    msg = Topicinit()
    msg.vehicle_token = data["vehicle_uuid"]
    msg.topic_uuid = data["uuid"]
    msg.topic_name = data["topic"]
    msg.topic_type = data["message_type"]

    msg.mqtt_id = agent.mqtt_id
    msg.frequency = int(agent.publish_frequency)
    msg.mode = agent.node.topic_streaming_mode

    msg.total_count = agent.total_topics
    msg.number_of_vehicles = vehicle_position + 1
    msg.prefix = agent.vehicles[vehicle_position]["prefix"]

    return msg


def parse_docking_routine(result):
    path = [[pose.pose.position.x, pose.pose.position.y] for pose in result.poses]
    rotation_angles = [[round(quaternion_to_euler(
        pose.pose.orientation.x, pose.pose.orientation.y, pose.pose.orientation.z, pose.pose.orientation.w
    ), 3)] for pose in result.poses]

    return path, rotation_angles



import math

from sentry_sdk import capture_exception


def quaternion_to_euler(x, y, z, w):
    # conversion between ros auaternation into euler angles
    try:
        t0 = +2.0 * (w * x + y * z)
        t1 = +1.0 - 2.0 * (x * x + y * y)
        roll = math.degrees(math.atan2(t0, t1))

        t2 = +2.0 * (w * y - z * x)
        t2 = +1.0 if t2 > +1.0 else t2
        t2 = -1.0 if t2 < -1.0 else t2
        pitch = math.degrees(math.asin(t2))

        t3 = +2.0 * (w * z + x * y)
        t4 = +1.0 - 2.0 * (y * y + z * z)
        yaw = math.degrees(math.atan2(t3, t4))

        if yaw < 0:
            yaw = yaw + 360
    except TypeError as e:
        capture_exception(TypeError)

        raise

    return yaw


def euler_to_quaternion(roll, pitch, yaw):
    try:
        # Conversion Between euler angles to quaternation
        qx = math.sin(roll / 2) * math.cos(pitch / 2) * math.cos(yaw / 2) - math.cos(
            roll / 2
        ) * math.sin(pitch / 2) * math.sin(yaw / 2)
        qy = math.cos(roll / 2) * math.sin(pitch / 2) * math.cos(yaw / 2) + math.sin(
            roll / 2
        ) * math.cos(pitch / 2) * math.sin(yaw / 2)
        qz = math.cos(roll / 2) * math.cos(pitch / 2) * math.sin(yaw / 2) - math.sin(
            roll / 2
        ) * math.sin(pitch / 2) * math.cos(yaw / 2)
        qw = math.cos(roll / 2) * math.cos(pitch / 2) * math.cos(yaw / 2) + math.sin(
            roll / 2
        ) * math.sin(pitch / 2) * math.sin(yaw / 2)

        rotation_quaternion = {"qx": qx, "qy": qy, "qz": qz, "qw": qw}
    except TypeError as e:
        capture_exception(TypeError)
        raise
    return rotation_quaternion



from geometry_msgs.msg import PoseStamped


class DockingActionServer:
    def __init__(self, name):
        self.plan = None
        self.recording = None
        self.action_name = name

    def callback_pose(self, msg):

        if self.recording:
            pose = PoseStamped()
            pose.header = msg.header
            x_coord = round(msg.pose.pose.position.x, 2)
            y_coord = round(msg.pose.pose.position.y, 2)

            try:
                if len(self.plan.poses) != 0:
                    x_coord_prev = self.plan.poses[-1].pose.position.x
                    y_coord_prev = self.plan.poses[-1].pose.position.y
                else:
                    x_coord_prev = 0.0
                    y_coord_prev = 0.0

                if x_coord != x_coord_prev and y_coord != y_coord_prev:
                    pose.pose.position.x = x_coord
                    pose.pose.position.y = y_coord
                    pose.pose.orientation.x = msg.pose.pose.orientation.x
                    pose.pose.orientation.y = msg.pose.pose.orientation.y
                    pose.pose.orientation.z = msg.pose.pose.orientation.z
                    pose.pose.orientation.w = msg.pose.pose.orientation.w

                    self.plan.poses.append(pose)

            except Exception as e:
                self.log_info(f"[DockingServer] Error in callback pose: {e}")

    def callback_recording(self, msg):
        self.recording = msg.data
